"""LayerInfo class.
"""
import logging
import weakref

from ...layers.base import Layer

LOGGER = logging.getLogger("napari.async")


class LayerInfo:
    """Information about one layer the ChunkLoader is tracking.

    Parameters
    ----------
    layer : Layer
        The layer we are loading chunks for.

    Attributes
    ----------
    layer_id : int
        The id of the layer.
    layer_ref : weakref
        Weak reference to the layer.

    Notes
    -----
    We store a weak reference because an in-progress request should not prevent
    a layer from being deleted. Meanwhile once a request has finished, we can
    de-reference to make sure the layer still exists.
    """

    def __init__(self, layer):
        self.layer_id: int = id(layer)
        self.layer_ref: weakref.ReferenceType = weakref.ref(layer)

    def get_layer(self) -> Layer:
        """Resolve our weakref to get the layer.

        Returns
        -------
        layer : Layer
            The layer for this ChunkRequest.
        """
        layer = self.layer_ref()
        if layer is None:
            LOGGER.debug(
                "LayerInfo.get_layer: layer %d was deleted", self.layer_id
            )
        return layer



"""ChunkLoader related utilities.
"""


def get_data_id(layer) -> int:
    """Return the data_id to use for this layer.

    Parameters
    ----------
    layer
        The layer to get the data_id from.

    Notes
    -----
    We use data_id rather than just the layer_id, because if someone
    changes the data out from under a layer, we do not want to use the
    wrong chunks.
    """
    data = layer.data
    if isinstance(data, list):
        assert data  # data should not be empty for image layers.
        return id(data[0])  # Just use the ID from the 0'th layer.

    return id(data)  # Not a list, just use it.



"""ChunkLoader: synchronous or asynchronous chunk loading
"""
import logging
import os
from concurrent.futures import CancelledError, Future, ThreadPoolExecutor
from typing import Dict, List, Optional

from ...types import ArrayLike
from ...utils.events import EmitterGroup
from ._info import LayerInfo
from ._request import ChunkKey, ChunkRequest

LOGGER = logging.getLogger("napari.async")


def _is_enabled(env_var) -> bool:
    """Return True if env_var is defined and not zero."""
    return os.getenv(env_var, "0") != "0"


def _chunk_loader_worker(request: ChunkRequest) -> ChunkRequest:
    """This is the worker thread that loads the array.

    We call np.asarray() in a worker because it might lead to IO or
    computation which would block the GUI thread.

    Parameters
    ----------
    request : ChunkRequest
        The request to load.
    """
    request.load_chunks()  # loads all chunks in the request
    return request


class ChunkLoader:
    """Loads chunks synchronously or asynchronously in worker threads.

    We cannot call np.asarray() in the GUI thread because it might block on
    IO or a computation. So the ChunkLoader calls np.asarray() in a worker
    thread if doing async loading.

    Attributes
    ----------
    synchronous : bool
        If True all requests are loaded synchronously.
    executor : ThreadPoolExecutor
        Our thread pool executor.
    futures : Dict[int, List[Future]]
        In progress futures for each layer (data_id).
    events : EmitterGroup
        We only signal one event: chunk_loaded.
    layer_map : Dict[int, LayerInfo]
        Stores a LayerInfo about each layer we are tracking.
    """

    def __init__(self):
        self.synchronous = not _is_enabled("NAPARI_ASYNC")
        self.executor = ThreadPoolExecutor(max_workers=6)
        self.futures: Dict[int, List[Future]] = {}
        self.events = EmitterGroup(
            source=self, auto_connect=True, chunk_loaded=None
        )
        self.layer_map: Dict[int, LayerInfo] = {}

    def get_info(self, layer_id: int) -> Optional[LayerInfo]:
        """Get LayerInfo for this layer or None."""
        return self.layer_map.get(layer_id)

    def create_request(
        self, layer, key: ChunkKey, chunks: Dict[str, ArrayLike]
    ) -> ChunkRequest:
        """Create a ChunkRequest for submission to load_chunk.

        Parameters
        ----------
        layer : Layer
            The layer that's requesting the data.
        key : ChunkKey
            The key for the request.
        chunks : Dict[str, ArrayLike]
            The arrays we want to load.
        """
        layer_id = key.layer_id

        # Add a LayerInfo if we don't already have one.
        if layer_id not in self.layer_map:
            self.layer_map[layer_id] = LayerInfo(layer)

        # Return the new request.
        return ChunkRequest(key, chunks)

    def load_chunk(self, request: ChunkRequest) -> Optional[ChunkRequest]:
        """Load the given request sync or async.

        Parameters
        ----------
        request : ChunkRequest
            Contains the array or arrays to load.

        Returns
        -------
        Optional[ChunkRequest]
            The ChunkRequest if it was satisfied otherwise None.

        Notes
        -----
        We return a ChunkRequest if we performed the load synchronously,
        otherwise we return None is indicating that an asynchronous load
        was intitiated. When the async load finishes the layer's
        on_chunk_loaded() will be called from the GUI thread.
        """
        if self._load_synchronously(request):
            return request

        # Clear any pending requests for this specific data_id.
        self._clear_pending(request.key.data_id)

        # Sumbit the request asynchronously.
        self._submit_async(request)

    def _load_synchronously(self, request: ChunkRequest) -> bool:
        """Return True if we loaded the request synchronously."""

        # If async is enabled, loaded non-ndarray request async.
        if not self.synchronous and not request.in_memory:
            return False  # load async

        # Load synchronously.
        request.load_chunks()
        return True

    def _submit_async(self, request: ChunkRequest) -> None:
        """Initiate an asynchronous load of the given request.

        Parameters
        ----------
        request : ChunkRequest
            Contains the arrays to load.
        """
        # Note about string formatting with logging: it's recommended to
        # use the oldest style of string formatting with logging. With
        # f-strings you'd pay the price of formatting the string even if
        # the log statement is disabled due to the log level, etc. In our
        # case the log will almost always be disabled unless debugging.
        # https://docs.python.org/3/howto/logging.html#optimization
        # https://blog.pilosus.org/posts/2020/01/24/python-f-strings-in-logging/
        LOGGER.debug("ChunkLoader._load_async: %s", request.key)

        # Submit the future, have it call ChunkLoader._done when done.
        future = self.executor.submit(_chunk_loader_worker, request)
        future.add_done_callback(self._done)

        # Store the future in case we need to cancel it.
        self.futures.setdefault(request.key.data_id, []).append(future)

    def _clear_pending(self, data_id: int) -> None:
        """Clear any pending requests for this data_id.

        Parameters
        ----------
        data_id : int
            Clear all requests associated with this data_id.
        """
        LOGGER.debug("ChunkLoader._clear_pending %d", data_id)

        # Get list of futures we submitted to the pool.
        future_list = self.futures.setdefault(data_id, [])

        # Try to cancel all futures in the list, but cancel() will return
        # False if the task already started running.
        num_before = len(future_list)
        future_list[:] = [x for x in future_list if x.cancel()]
        num_after = len(future_list)
        num_cleared = num_before - num_after

        # Delete the list entirely if empty
        if num_after == 0:
            del self.futures[data_id]

        # Log what we did.
        if num_before == 0:
            LOGGER.debug("ChunkLoader.clear_pending: empty")
        else:
            LOGGER.debug(
                "ChunkLoader.clear_pending: %d of %d cleared -> %d remain",
                num_cleared,
                num_before,
                num_after,
            )

    @staticmethod
    def _get_request(future: Future) -> Optional[ChunkRequest]:
        """Return the ChunkRequest for this future.

        Parameters
        ----------
        future : Future
            Get the request from this future.

        Returns
        -------
        Optional[ChunkRequest]
            The ChunkRequest or None if the future was cancelled.
        """
        try:
            # Our future has already finished since this is being
            # called from Chunk_Request._done(), so result() will
            # never block.
            return future.result()
        except CancelledError:
            LOGGER.debug("ChunkLoader._done: cancelled")
            return None

    def _done(self, future: Future) -> None:
        """Called when a future finishes with success or was cancelled.

        Parameters
        ----------
        future : Future
            The future that finished or was cancelled.

        Notes
        -----
        This method may be called in the worker thread. The
        concurrent.futures documentation very intentionally does not
        specify which thread the future's done callback will be called in,
        only that it will be called in some thread in the current process.
        """
        try:
            request = self._get_request(future)
        except ValueError:
            return  # Pool not running, app exit in progress.

        if request is None:
            return  # Future was cancelled, nothing to do.

        LOGGER.debug("ChunkLoader._done: %s", request.key)

        # Lookup this Request's LayerInfo.
        info = self._get_layer_info(request)

        # Resolve the weakref.
        layer = info.get_layer()

        if layer is None:
            return  # Ignore chunks since layer was deleted.

        # Fire event to tell QtChunkReceiver to forward this chunk to its
        # layer in the GUI thread.
        self.events.chunk_loaded(layer=layer, request=request)

    def _get_layer_info(self, request: ChunkRequest) -> LayerInfo:
        """Return LayerInfo associated with this request or None.

        Parameters
        ----------
        request : ChunkRequest
            Return Layer_info for this request.

        Raises
        ------
        KeyError
            If the layer is not found.
        """
        layer_id = request.key.layer_id

        # Raises KeyError if not found. This should never happen because we
        # add the layer to the layer_map in ChunkLoader.create_request().
        return self.layer_map[layer_id]

    def on_layer_deleted(self, layer):
        """The layer was deleted, delete it from our map.

        Layer
            The layer that was deleted.
        """
        try:
            del self.layer_map[id(layer)]
        except KeyError:
            pass  # We weren't tracking that layer yet.


# Global instance
chunk_loader = ChunkLoader()



"""ChunkRequest is passed to ChunkLoader.load_chunks().
"""
import logging
from typing import Optional, Tuple

import numpy as np

from ...layers.base.base import Layer
from ...types import ArrayLike, Dict
from ._utils import get_data_id

LOGGER = logging.getLogger("napari.async")

# We convert slices to tuple for hashing.
SliceTuple = Tuple[Optional[int], Optional[int], Optional[int]]


def _flatten(indices):
    result = []
    for x in indices:
        if isinstance(x, slice):
            result.extend([x.start, x.stop, x.step])
        else:
            result.append(x)
    return tuple(result)


class ChunkKey:
    """The key for one single ChunkRequest.

    Parameters
    ----------
    layer : Layer
        The layer to load data for.
    indices : Indices
        The indices to load from the layer.

    Attributes
    ----------
    layer_id : int
        The id of the layer making the request.
    data_level : int
        The level in the data (for multi-scale).
    indices : Tuple[Optional[slice], ...]
        The indices of the slice.
    key : Tuple
        The combined key, all the identifiers together.
    """

    def __init__(self, layer: Layer, indices: Tuple[Optional[slice], ...]):
        self.layer_id = id(layer)
        self.data_id = get_data_id(layer)
        self.data_level = layer._data_level
        self.indices = indices

        combined = (
            self.layer_id,
            self.data_id,
            self.data_level,
            _flatten(self.indices),
        )
        self.key = hash(combined)

    def __str__(self):
        return (
            f"layer_id={self.layer_id} data_id={self.data_id} "
            f"data_level={self.data_level} indices={self.indices}"
        )

    def __eq__(self, other):
        return self.key == other.key


class ChunkRequest:
    """A request asking the ChunkLoader to load one or more arrays.

    Parameters
    ----------
    key : ChunkKey
        The key of the request.
    chunks : Dict[str, ArrayLike]
        The chunk arrays we need to load.

    Attributes
    ----------
    key : ChunkKey
        The key of the request.
    chunks : Dict[str, ArrayLike]
        The chunk arrays we need to load.
    """

    def __init__(self, key: ChunkKey, chunks: Dict[str, ArrayLike]):
        # Make sure chunks dict is what we expect.
        for chunk_key, array in chunks.items():
            assert isinstance(chunk_key, str)
            assert array is not None

        self.key = key
        self.chunks = chunks

    @property
    def in_memory(self) -> bool:
        """True if all chunks are ndarrays."""
        return all(isinstance(x, np.ndarray) for x in self.chunks.values())

    def load_chunks(self):
        """Load all of our chunks now in this thread."""
        for key, array in self.chunks.items():
            loaded_array = np.asarray(array)
            self.chunks[key] = loaded_array

    def transpose_chunks(self, order: tuple) -> None:
        """Transpose all our chunks.

        Parameters
        ----------
        order
            Transpose the chunks into this order.
        """
        for key, array in self.chunks.items():
            self.chunks[key] = array.transpose(order)

    @property
    def image(self):
        """The image chunk if we have one or None."""
        return self.chunks.get('image')

    @property
    def thumbnail_source(self):
        """The chunk to use as the thumbnail_source or None."""
        try:
            return self.chunks['thumbnail_source']
        except KeyError:
            # No thumbnail_source so return the image instead. For single-scale
            # we use the image as the thumbnail_source.
            return self.chunks.get('image')



"""ImageSlice class.
"""
import logging
from typing import Callable, Optional

import numpy as np

from ...components.chunk import ChunkKey, ChunkRequest, chunk_loader
from ...types import ArrayLike
from ._image_view import ImageView

LOGGER = logging.getLogger("napari.async")


class ImageSlice:
    """The slice of the image that we are currently viewing.

    Parameters
    ----------
    image : ArrayLike
        The initial image used as the image and the thumbnail source.
    image_converter : Callable[[ArrayLike], ArrayLike]
        ImageView uses this to convert from raw to viewable.
    rgb : bool
        True if the image is RGB format. Otherwise its RGBA.

    Attributes
    ----------
    image : ImageView
        The main image for this slice.
    thumbnail : ImageView
        The source image used to compute the smaller thumbnail image.
    rgb : bool
        Is the image in RGB or RGBA format.
    current_key : Optional[ChunkKey]
        The ChunkKey we are currently showing or which is loading.
    loaded : bool
        Has the data for this slice been loaded yet.
    """

    def __init__(
        self,
        image: ArrayLike,
        image_converter: Callable[[ArrayLike], ArrayLike],
        rgb: bool = False,
    ):
        LOGGER.debug("ImageSlice.__init__")
        self.image: ImageView = ImageView(image, image_converter)
        self.thumbnail: ImageView = ImageView(image, image_converter)
        self.rgb = rgb

        # We're showing nothing to start.
        self.current_key: Optional[ChunkKey] = None

        # With async there can be a gap between when the ImageSlice is
        # created and the data is actually loaded. However initialize
        # as True in case we aren't even doing async loading.
        self.loaded = True

    def set_raw_images(
        self, image: ArrayLike, thumbnail_source: ArrayLike
    ) -> None:
        """Set the image and its thumbnail.

        If floating point / grayscale then clip to [0..1].

        Parameters
        ----------
        image : ArrayLike
            Set this as the main image.
        thumbnail : ArrayLike
            Derive the thumbnail from this image.
        """
        if self.rgb and image.dtype.kind == 'f':
            image = np.clip(image, 0, 1)
            thumbnail_source = np.clip(thumbnail_source, 0, 1)
        self.image.raw = image
        self.thumbnail.raw = thumbnail_source
        self.loaded = True

    def load_chunk(self, request: ChunkRequest) -> Optional[ChunkRequest]:
        """Load the requested chunk asynchronously.

        Parameters
        ----------
        request : ChunkRequest
            Load this chunk sync or async.
        """
        LOGGER.debug("ImageSlice.load_chunk: %s", request.key)

        if self.current_key is not None and self.current_key == request.key:
            # We are already showing this slice, or its being loaded
            # asynchronously.
            return None

        # Now "showing" this slice, even if it hasn't loaded yet.
        self.current_key = request.key
        self.loaded = False

        # This will return a satisfied request if the load was synchronous.
        # If it returns None that means a request was queued and it will be
        # loaded in a worker thread or process.
        return chunk_loader.load_chunk(request)

    def on_chunk_loaded(self, request: ChunkRequest) -> bool:
        """Chunk was loaded, show this new data.

        Parameters
        ----------
        request : ChunkRequest
            The chunk request that was loaded in a worker thread or process.

        Return
        ------
        bool
            False if the chunk was for the wrong slice and was not used.
        """
        # Is this the chunk we last requested?
        if not self.current_key == request.key:
            # Probably we are scrolling through slices and we are no longer
            # showing this slice, so drop it. It should have been added
            # to the cache so the load was not totally wasted.
            LOGGER.debug("ImageSlice.chunk_loaded: reject %s", request.key)
            return False

        LOGGER.debug("ImageSlice.chunk_loaded: accept %s", request.key)

        # Display the newly loaded data.
        self.set_raw_images(request.image, request.thumbnail_source)
        return True



import numpy as np
import pytest


@pytest.mark.filterwarnings("ignore:data shape:UserWarning")
def test_big_2D_image(make_test_viewer):
    """Test big 2D image with axis exceeding max texture size."""
    viewer = make_test_viewer()

    shape = (20_000, 10)
    data = np.random.random(shape)
    layer = viewer.add_image(data, multiscale=False)
    visual = viewer.window.qt_viewer.layer_to_visual[layer]
    assert visual.node is not None
    if visual.MAX_TEXTURE_SIZE_2D is not None:
        s = np.ceil(np.divide(shape, visual.MAX_TEXTURE_SIZE_2D)).astype(int)
        assert np.all(layer._transforms['tile2data'].scale == s)


@pytest.mark.filterwarnings("ignore:data shape:UserWarning")
def test_big_3D_image(make_test_viewer):
    """Test big 3D image with axis exceeding max texture size."""
    viewer = make_test_viewer(ndisplay=3)

    shape = (5, 10, 3_000)
    data = np.random.random(shape)
    layer = viewer.add_image(data, multiscale=False)
    visual = viewer.window.qt_viewer.layer_to_visual[layer]
    assert visual.node is not None
    if visual.MAX_TEXTURE_SIZE_3D is not None:
        s = np.ceil(np.divide(shape, visual.MAX_TEXTURE_SIZE_3D)).astype(int)
        assert np.all(layer._transforms['tile2data'].scale == s)
